#!/usr/bin/env python
import numpy as np
import pandas as pd
import paths as pt
from tools import file_reader, file_writer, explainer
from utility import metrics
from sklearn.metrics import accuracy_score, precision_score
from sklearn.metrics import recall_score, roc_auc_score
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import train_test_split
import xgboost as xgb

CASES = ["Complete", "Compliance", "Fall", "Risk"]

def main(dataset_version : str = 'emb'):
    for case in CASES:
        if case == "Complete":
            if dataset_version == 'ohe':
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'complete_count.csv')
            else:
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'complete_emb.csv')
            model_dir = pt.COMPLETE_XGB_DIR
            target_name = "Complete"
        elif case == "Compliance":
            if dataset_version == 'ohe':
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'compliance_count.csv')
            else:
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'compliance_emb.csv')
            model_dir = pt.COMPLIANCE_XGB_DIR
            target_name = "Compliance"
        elif case == "Fall":
            if dataset_version == 'ohe':
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_count.csv')
            else:
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_emb.csv')
            model_dir = pt.FALL_XGB_DIR
            target_name = "Fall"
        else:
            if dataset_version == 'ohe':
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'risk_count.csv')
            else:
                df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'risk_emb.csv')
            model_dir = pt.RISK_XGB_DIR
            target_name = "Risk"
    
        df = df.sample(frac=1, random_state=0).reset_index(drop=True)
        
        X = df.drop([target_name], axis=1)
        y = df[target_name]
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3,
                                                            stratify=y, random_state=0)
        
        neg, pos = np.bincount(y)
        scale_pos_weight = neg / pos
    
        params = {"n_estimators": 400,
                "objective": "binary:logistic",
                "scale_pos_weight": scale_pos_weight,
                "use_label_encoder": False,
                "learning_rate": 0.07,
                "eval_metric": "logloss",
                "seed": 0
        }
                
        model = xgb.XGBClassifier(**params)
        skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
    
        y_valid_pred = 0*y
        valid_acc, valid_pre, valid_recall, valid_roc_auc = list(), list(), list(), list()
        for train_index, valid_index in skf.split(X_train, y_train):
            X_train_split, X_valid_split = X_train.iloc[train_index,:], X_train.iloc[valid_index,:]
            y_train_split, y_valid_split = y_train.iloc[train_index], y_train.iloc[valid_index]

            optimize_rounds = True
            early_stopping_rounds = 200
            if optimize_rounds:
                eval_set=[(X_valid_split, y_valid_split)]
                fit_model = model.fit(X_train_split, y_train_split,
                                      eval_set=eval_set,
                                      eval_metric=metrics.gini_xgb,
                                      early_stopping_rounds=early_stopping_rounds,
                                      verbose=False)
            else:
                fit_model = model.fit(X_train_split, y_train_split)
        
            pred = fit_model.predict_proba(X_valid_split)[:,1]
            y_valid_pred.iloc[valid_index] = pred
            
            y_valid_scores = (y_valid_pred.iloc[valid_index] > 0.5)
            valid_acc.append(accuracy_score(y_valid_split, y_valid_scores))
            valid_pre.append(precision_score(y_valid_split, y_valid_scores))
            valid_recall.append(recall_score(y_valid_split, y_valid_scores))
            valid_roc_auc.append(roc_auc_score(y_valid_split, y_valid_pred.iloc[valid_index]))
                
        y_pred = model.predict(X_test)
        y_proba = model.predict_proba(X_test)[:,1]
        file_writer.write_cm_plot(y_test, y_pred, pt.REPORTS_PLOTS_DIR,
                                f'{case.lower()}_xgb_cm.pdf', case)
        file_writer.write_joblib(model, model_dir, f'{case.lower()}_xgboost.joblib')
        
        print(f"Scores for {case} XGBoost model:")
        print(f"Accuracy: {np.around(accuracy_score(y_test, y_pred), decimals=3)}")
        print(f"Precision: {np.around(precision_score(y_test, y_pred), decimals=3)}")
        print(f"Recall: {np.around(recall_score(y_test, y_pred), decimals=3)}")
        print(f"ROC AUC: {np.around(roc_auc_score(y_test, y_proba), decimals=3)}\n")
        
        feature_names = X.columns
        shap_explainer, shap_values = explainer.get_shap_tree_explainer(model, X_test=X_test)
        
        importance_df  = pd.DataFrame()
        importance_df['feature'] = feature_names
        importance_df['shap_values'] = np.around(abs(np.array(shap_values)[:,:]).mean(0), decimals=3)
        importance_df['feat_imp'] = np.around(model.feature_importances_, decimals=3)
        feat_importance_df_shap = importance_df.groupby('feature').mean().sort_values('shap_values',
                                                                                       ascending=False)
        feat_importance_df_shap = feat_importance_df_shap.reset_index()
        file_writer.write_csv(feat_importance_df_shap, pt.REPORTS_DIR, f"{case} best features.csv")
        file_name_sum = f'{case.lower()}_shap_summary'
        file_name_exp = f'{case.lower()}_shap_row_0'
        explainer.make_shap_summary_plot(shap_values, X_test, feature_names,
                                         file_name_sum, model_dir,
                                         plot_type=None)
        explainer.make_shap_summary_plot(shap_values, X_test, feature_names,
                                     file_name_sum + '_bar', model_dir,
                                     plot_type="bar")
    
if __name__ == '__main__':
    main()
