from pandas.core.arrays import boolean
from tools import file_reader
import paths as pt
import pandas as pd
import numpy as np
import datetime
from pathlib import Path
import abc

class BaseParser(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def parse_status_set(self, file_name, path):
        """Parse the DiGiRehab StatusSet data set"""
    
    @abc.abstractmethod
    def parse_training_cancelled(self, file_name, path):
        """Parse the DiGiRehab Training Cancelled data set"""
    
    @abc.abstractmethod
    def parse_screening_content(self, file_name, path):
        """Parse the DiGiRehab Screening Content data set"""

    @abc.abstractmethod
    def parse_training_done(self, file_name, path):
        """Parse the DiGiRehab Training Done data set"""
    
    @abc.abstractmethod
    def parse_assistive_aids(self, file_names, path):
        """Parse the DiGiRehab Assistive Aids data set"""
        
    @abc.abstractmethod
    def parse_clusters(self, file_name, path):
        """Parse the generated clusters data set"""
    
    @abc.abstractmethod    
    def parse_iso_classes(self, file_name, path):
        """Parse the ISO classes data set"""
    
    @abc.abstractmethod
    def parse_patient_data(self, file_name, path):
        """Parse the DiGiRehab Patient data set"""
    
    @abc.abstractmethod
    def parse_fall_data(self, file_name, path):
        """Parse the DiGiRehab fall data set"""
    
class Parser2020(BaseParser):
    def assert_datetime(self, entry: str) -> bool:
        return pd.to_datetime(entry, format='%d-%m-%Y', errors='coerce') is not pd.NaT
    
    def parse_status_set(self, file_name: str, path: Path) -> pd.DataFrame:
        index_col = 'BorgerID'
        converters = {index_col: str} # pass as string to avoid float conversion
        ss_df = file_reader.read_excelfile(path, file_name, converters=converters)

        data_columns = ['CitizenId', 'ChangeDate', 'Status']
        
        df = pd.DataFrame(columns=data_columns)
        index_row = 6
        for _, row in ss_df.iterrows():
            if not pd.isna(row[index_row]):
                list_of_entries = [str.split(line, ';') for line in str.split(row[index_row], '#')]
                for entry in list(filter(None, list_of_entries)):
                    entry.insert(0, row[index_col])
                    entry[-1] = entry[-1].replace('[', '').replace(']', '')
                    if not self.assert_datetime(entry[1]): # Handle a case where time is missing
                        entry[1] = np.nan
                        entry[2] = 'AutoTerminated'
                    entry_data = pd.DataFrame([entry[:3]], columns=data_columns)
                    df = df.append(entry_data, ignore_index=True)
        
        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        df['ChangeDate'] = pd.to_datetime(df['ChangeDate'], format='%d-%m-%Y')
        df['Status'] = pd.Series.astype(df['Status'], dtype=str)
        
        return df
    
    def parse_training_cancelled(self, file_name: str, path: Path) -> pd.DataFrame:
        index_col = 'BorgerID'
        converters = {index_col: str} # pass as string to avoid float conversion
        tc_df = file_reader.read_excelfile(path, file_name, converters=converters)
        
        data_columns = ['CitizenId', 'RatingDate', 'RatingScore']
        
        df = pd.DataFrame(columns=data_columns)
        for _, row in tc_df.iterrows():
            if not pd.isna(row[5]):
                list_of_entries = [str.split(line, ';') for line in str.split(row[5], '#')]
                for entry in list(filter(None, list_of_entries)):
                    entry.insert(0, row[index_col])
                    entry[-1] = entry[-1].replace('[', '').replace(']', '')
                    entry_data = pd.DataFrame([entry[:3]], columns=data_columns)
                    df = df.append(entry_data, ignore_index=True)

        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        df['RatingDate'] = pd.to_datetime(df['RatingDate'], format='%d-%m-%Y')
        df['RatingScore'] = pd.Series.astype(df['RatingScore'], dtype=int)
        
        return df

    def parse_screening_content(self, file_name: str, path: Path) -> pd.DataFrame:
        index_col = 'BorgerID'
        converters = {index_col: str} # pass as string to avoid float conversion
        sc_df = file_reader.read_excelfile(path, file_name, converters=converters)
        
        data_columns = ['CitizenId', 'ScreeningDate', 'NeedForHelpScore',
                    'NeedForHelpReason', 'PhysicalStrengthScore',
                    'PhysicalStrengthReason', 'ExerciseContent']
        
        df = pd.DataFrame(columns=data_columns)
        for _, row in sc_df.iterrows():
            if not pd.isna(row[3]):
                list_of_entries = [str.split(line, ';') for line in str.split(row[3], '#')]
                for entry in list(filter(None, list_of_entries)):
                    entry.insert(0, row[index_col])
                    entry[-1] = entry[-1].replace('[', '').replace(']', '')
                    entry_data = pd.DataFrame([entry], columns=data_columns)
                    df = df.append(entry_data, ignore_index=True)
        
        # Handle missing values
        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        df['NeedForHelpScore'] = df['NeedForHelpScore'].fillna(0)
        df['PhysicalStrengthScore'] = df['PhysicalStrengthScore'].fillna(0)
        df['ExerciseContent'] = df['ExerciseContent'].fillna('0')
        
        df['ScreeningDate'] = pd.to_datetime(df['ScreeningDate'], format='%d-%m-%Y')
        df['NeedForHelpScore'] = pd.Series.astype(df['NeedForHelpScore'], dtype=int)
        df['NeedForHelpReason'] = pd.Series.astype(df['NeedForHelpReason'], dtype=str)
        df['PhysicalStrengthScore'] = pd.Series.astype(df['PhysicalStrengthScore'], dtype=int)
        df['PhysicalStrengthReason'] = pd.Series.astype(df['PhysicalStrengthReason'], dtype=str)
        df['ExerciseContent'] = pd.Series.astype(df['ExerciseContent'], dtype=str)
        
        return df

    def parse_training_done(self, file_name: str, path: Path) -> pd.DataFrame:
        index_col = 'BorgerID'
        converters = {index_col: str} # pass as string to avoid float conversion
        td_df = file_reader.read_excelfile(path, file_name, converters=converters)
        
        data_columns = ['CitizenId', 'RatingDate', 'RatingScore', 'Gender', 'BirthYear']
        
        df = pd.DataFrame(columns=data_columns)
        for _, row in td_df.iterrows():
            if not pd.isna(row[4]):
                list_of_entries = [str.split(line, ";") for line in str.split(row[4], "#")]
                for entry in list(filter(None, list_of_entries)):
                    if len(entry) > 2:
                        entry.insert(0, row[index_col])
                        entry[-1] = entry[-1].replace('[', '').replace(']', '')
                        data_entry = pd.DataFrame([entry[:3] + [row[1]] + [row[2]]], columns=data_columns)
                        df = df.append(data_entry, ignore_index=True)
        
        # Handle missing values
        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        df['BirthYear'] = df['BirthYear'].fillna(0)
        
        df['RatingDate'] = pd.to_datetime(df['RatingDate'], format='%d-%m-%Y')
        df['RatingScore'] = pd.Series.astype(df['RatingScore'], dtype=int)
        df['Gender'] = pd.Series.astype(df['Gender'], dtype=str)
        df['BirthYear'] = pd.Series.astype(df['BirthYear'], dtype=int)
        return df
    
    def parse_assistive_aids(self, file_name: str,
                             path: Path, n_sheets: int=6) -> pd.DataFrame:
        converters = {'BorgerID': str,
                      'HJAELPEMIDDELHMINUMMER': str,
                      'HJAELPEMIDDELISOKLASSE': str} # pass as str to avoid int conversion
        
        df = file_reader.read_excelfile_sheets(path, file_name, n_sheets, converters)
        df = df.rename(columns={'BorgerID': 'CitizenId',
                                'Gender': 'Gender',
                                'Birth Year': 'BirthYear',
                                'HJAELPEMIDDELHMINUMMER': 'DevHMINumber',
                                'HJAELPEMIDDELHMINAVN': 'DevHMIName',
                                'HJAELPEMIDDELISOKLASSE': 'DevISOClass',
                                'HJAELPEMIDDELSERIENUMMER': 'DevSerial',
                                'LOVPARAGRAFNAVN': 'LawParagraph',
                                'UDLEVERINGSDATO': 'LendDate',
                                'INDLEVERINGSDATO': 'ReturnDate',
                                'PRIS': 'Price'})
        
        # Handle missing values
        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        df['LawParagraph'] = df['LawParagraph'].fillna(0)
        df['Price'] = df['Price'].fillna(0)
        
        df['Gender'] = pd.Series.astype(df['Gender'], dtype=str)
        df['DevHMINumber'] = pd.Series.astype(df['DevHMINumber'], dtype=str)
        df['DevHMIName'] = pd.Series.astype(df['DevHMIName'], dtype=str)
        df['DevISOClass'] = pd.Series.astype(df['DevISOClass'], dtype=str)
        df['DevSerial'] = pd.Series.astype(df['DevSerial'], dtype=str)
        df['LawParagraph'] = pd.Series.astype(df['LawParagraph'], dtype=str)
        df['LendDate'] = pd.to_datetime(df['LendDate'], format='%d-%m-%Y')
        df['ReturnDate'] = pd.to_datetime(df['ReturnDate'], format='%Y-%m-%d', errors = 'coerce')
        df['Price'] = pd.Series.astype(df['Price'], dtype=str)
        return df
    
    def parse_clusters(self, file_name: str, path: Path) -> pd.DataFrame:
        cluster_file = Path.joinpath(path, file_name)
        df = pd.read_csv(cluster_file)
        return df
    
    def parse_iso_classes(self, file_name: str, path: Path) -> pd.DataFrame:
        isoclass_file = Path.joinpath(path, file_name)
        df = pd.read_csv(isoclass_file,
                        header=None,
                        usecols=[0, 1, 2],
                        names=['DevISOClass', 'GroupSize', 'Description'],
                        converters={i: str for i in range(0, 10000)})
        return df

    def parse_fall_data(self, file_name: str, path: Path) -> pd.DataFrame:
        converters = {'BorgerId': str} # pass as string to avoid float conversion
        df = file_reader.read_excelfile(path, file_name, converters)
        df = df.rename(columns={'BorgerId': 'CitizenId',
                                'Birth year': 'BirthYear',
                                'Dato': 'Date',
                                'Spørgmål lang': 'Question',
                                'Svar lang': 'Answer'})
        
        # Handle missing values
        df.replace(to_replace='', value=np.nan, regex=True, inplace=True)
        
        df['CitizenId'] = pd.Series.astype(df['CitizenId'], dtype=str)
        df['Gender'] = pd.Series.astype(df['Gender'], dtype=str)
        df['BirthYear'] = pd.Series.astype(df['BirthYear'], dtype=int)
        df['Date'] = pd.to_datetime(df['Date'], format='%d-%m-%Y')
        df['Question'] = pd.Series.astype(df['Question'], dtype=str)
        df['Answer'] = pd.Series.astype(df['Answer'], dtype=str)
            
        return df
        
    def parse_patient_data(self, file_name: str, path: Path) -> pd.DataFrame:
        raise NotImplementedError