import pandas as pd
import numpy as np
import paths as pt
import yaml
from pathlib import Path
from pandas._libs.tslibs import Timestamp
from typing import Tuple
from utility import data_dto
from pandas.tseries.offsets import DateOffset
from abc import ABC, abstractmethod
    
def make_risk_target(df: pd.DataFrame,
                      cura_falls: pd.DataFrame,
                      settings: dict):
    # Get Cura and DigiRehab falls, respectively
    cura_falls = cura_falls[['CitizenId', 'Date']]
    digi_falls = df[['CitizenId', 'NeedsReason', 'PhysicsReason', 'EndDate']].fillna('Ingen')
    digi_falls = digi_falls[digi_falls['NeedsReason'].str.contains("Fald/uheld")
                            | digi_falls['PhysicsReason'].str.contains("Fald/uheld")]
    
    # Make target by annotating falls in the risk period
    risk_period = settings['risk_period_months']
    df['Risk'] = df[['CitizenId', 'EndDate']].apply(lambda x:
        annotate_falls(x, digi_falls, cura_falls, risk_period), axis=1)
    
    return df
  
def make_compliance_target(df: pd.DataFrame, settings: dict) -> pd.DataFrame:
    df = accumulate_screenings(df, settings)
    
    # Set first screening as baseline
    df.loc[df['NumberScreening'] ==  0, 'Baseline'] = 1
    
    # Calculate if citizens reached compliance in their sessions
    grp_compliance = df.groupby(['CitizenId', 'NumberSplit', 'NumberSession'])['GotComplianceInSession']
    df['Compliance'] = grp_compliance.transform(lambda x: 1 if np.max(x) > 0 else 0)
    df = df.astype({"Compliance": int})
    
    # Calculate if citizens completed their sessions
    grp_complete = df.groupby(['CitizenId', 'NumberSplit', 'NumberSession'])['HasCompletedSession']
    df['Complete'] = grp_complete.transform(lambda x: 1 if np.max(x) > 0 else 0)
    df = df.astype({"Complete": int})

    # Select only the baseline screenings
    df = df.loc[df['Baseline'] == 1]
    
    # Select only those that complete
    df = df.loc[df['Complete'] == 1]
    
    # Reset index
    df = df.reset_index()

    return df
    
def make_complete_target(df: pd.DataFrame, settings: dict) -> pd.DataFrame:
    df = accumulate_screenings(df, settings)
    
    # Set first screening as baseline
    df.loc[df['NumberScreening'] == 0, 'Baseline'] = 1
    
    # Calculate if citizens complete their sessions
    grp = df.groupby(['CitizenId', 'NumberSplit', 'NumberSession'])['HasCompletedSession']
    df['Complete'] = grp.transform(lambda x: 1 if np.max(x) > 0 else 0)
    df = df.astype({"Complete": int})

    # Select only the baseline screenings
    df = df.loc[df['Baseline'] == 1]

    # Reset index
    df = df.reset_index()

    return df

def make_fall_target(df, settings):    
    df = accumulate_screenings(df, settings)

    # Set first screening as baseline
    df.loc[df['NumberScreening'] == 0, 'Baseline'] = 1

    # Calculate if citizens fall during their sessions
    grp = df.groupby(['CitizenId', 'NumberSplit', 'NumberSession'])['HasFallenInSession']
    df['Fall'] = grp.transform(lambda x: 1 if np.max(x) > 0 else 0)
    df = df.astype({"Fall": int})

    # Select only the baseline screenings
    df = df.loc[df['Baseline'] == 1]

    # Reset index
    df = df.reset_index()

    return df

def annotate_falls(row, digi_db, cura_db, risk_period):
    citizen_id = row['CitizenId']
    current_date = pd.Timestamp(row['EndDate'])
    end_date = current_date + DateOffset(months=risk_period)
    
    digi_db['EndDate'] = pd.to_datetime(digi_db['EndDate'])
    cura_db['Date'] = pd.to_datetime(cura_db['Date'])
    
    timespan_digi_falls = digi_db.loc[(digi_db['EndDate'] >= current_date)
                                      & (digi_db['CitizenId'] == citizen_id)
                                      & (digi_db['EndDate'] <= end_date)]
    
    timespan_cura_falls = cura_db.loc[(cura_db['Date'] >= current_date)
                                    & (cura_db['CitizenId'] == citizen_id)
                                    & (cura_db['Date'] <= end_date)]
    
    if len(timespan_digi_falls) > 0 or len(timespan_cura_falls) > 0:
        return 1
    return 0

def accumulate_screenings(df: pd.DataFrame, settings: dict) -> pd.DataFrame:
    for group_name, _ in df.groupby(['CitizenId', 'NumberSplit']):
        number_session = 0
        cumsum_weeks = 0
        cumsum_training = 0
        citizen_df = df.loc[df['CitizenId'] == group_name[0]]
        items_weeks = citizen_df['NumberWeeks'].iteritems()
        items_training = citizen_df['NumberTraining'].iteritems()
        needs_reason = citizen_df['NeedsReason'].iteritems()
        physics_reason = citizen_df['PhysicsReason'].iteritems()
        has_fall_risk = citizen_df['HasFallRisk'].iteritems()
        number_falls = citizen_df['NumberFalls'].iteritems()
        mean_evaluation = citizen_df['MeanEvaluation'].iteritems()
        for (row_week, row_train, row_needs_reason, row_physics_reason,
             row_has_fall_risk, row_number_falls, row_mean_evaluation) in zip(
            items_weeks, items_training, needs_reason, physics_reason,
            has_fall_risk, number_falls, mean_evaluation):
            df.loc[row_week[0], 'NumberSession'] = number_session
            cumsum_weeks += row_week[1]
            cumsum_training += row_train[1]
            if (row_needs_reason[1] == 'Fald/uheld') | (row_physics_reason[1] == 'Fald/uheld') \
                | (row_number_falls[1] > 0) | row_has_fall_risk[1] == True:
                df.loc[row_needs_reason[0], 'HasFallenInSession'] = 1
            else:
                df.loc[row_needs_reason[0], 'HasFallenInSession'] = 0
            if cumsum_weeks >= settings['threshold_weeks'] and cumsum_training >= settings['threshold_training']:
                cumsum_weeks = 0
                cumsum_training = 0
                number_session += 1
                df.loc[row_week[0], 'HasCompletedSession'] = 1
                df.loc[row_week[0], 'Baseline'] = 1
                if (row_mean_evaluation[1] > 4):
                    df.loc[row_week[0], 'GotComplianceInSession'] = 1
                else:
                    df.loc[row_week[0], 'GotComplianceInSession'] = 0
            else:
                df.loc[row_week[0], 'HasCompletedSession'] = 0
                df.loc[row_week[0], 'Baseline'] = 0
                df.loc[row_week[0], 'GotComplianceInSession'] = 0
    return df